# Additional Material for: A Case Study on Sampling Strategies for Evaluating Neural Sequential Item Recommendation Models
This project contains ASME config files that can be used to train and evaluate the models. 
Additionally, we provide code for creating the tables and figures that we use in the paper from the evaluation results.

We also provide a snapshot of our evaluation framework.
Please note that we will release a first version of the entire framework soon and the config files then be updated accordingly.

Additionally, we provide checkpoints and evaluation results for your convenience.
You can find the download links below. 

If something does not work as expected, please do not hesitate to contact me: [dallmann@informatik.uni-wuerzburg.de](mailto:dallmann@informatik.uni-wuerzburg.de)


# Downloads

* [ASME Snapshot](archives/asme.tar.gz)
* [Model Checkpoints](https://professor-x.de/papers/metrics-sampling-eval/checkpoints.tar.gz)
* [Evaluation Results](archives/results.tar.gz)

# Instructions

## Install asme
All models used in the paper can be run using the ASME framework. Therefore, ASME has to be installed to reproduce any of the results. <br/>
In order to install ASME extract the provided tar-archive into a folder called `asme`. 
Installing ASME requires [Poetry](https://python-poetry.org/). 
After installing Poetry, you can follow the steps below to obtain a local installation/container image of ASME:

### Local Installation
 After installing poetry, you can install ASME and its dependencies via:
```bash
cd asme && poetry install
```
### Installation into a container
In order to build a container with ASME installed, you can use `docker`/`podman` as follows:
```bash
bash k8s/release/build.sh
```
The resulting image is called `localhost/asme:0.5.0`
### Verifying your Installation
In order to verify your local installation, simply run 
```bash
poetry shell
```
inside the `asme` directory to activate the virtual environment, followed by 
```bash
asme
```
If using a container, simply execute
```bash
docker run -it localhost/asme:0.5.0
```
In both cases, ASME was installed successfully, if the `asme` command was found.
## Train models
The following section assumes that you are located in the `asme` folder have already activate the virtual environment via `poetry shell` (either locally or inside of the container). <br/>

### Preparing the datasets
All datasets found in the paper can be downloaded and preprocessed automatically using `asme`.
Here are the commands to download and preprocess a specific dataset:


#### Amazon Beauty

```bash
asme-datasets pre_process amazon --category beauty
```

#### Amazon Games

```bash
asme-datasets pre_process amazon --category games
```

#### Movielens 1m (ML-1m)
```bash
asme-datasets pre_process movielens ml-1m --min-user-feedback=5 --min-item-feedback=5
```

#### Movielens 20m (ML-20m)

```bash
asme-datasets pre_process movielens ml-20m
```

#### Steam

```bash
asme-datasets pre_process steam
```


The datasets are then stored at `./dataset/NAME`. When using a container you should mount a persistent volume to this directory to avoid repeating the processing upon every run. Also note, that you will need a shell in the container to download datasets. To do so, execute:
```bash
docker run -it --entrypoint /bin/bash localhost/asme:0.5.0
``` 
The other datasets follow the same command pattern, just substitute the dataset name / variant, e.g. by `amazon beauty` or `steam`. 
Refer to the `--help` command for further information.
### Preparing the training environment
In order to be independent from specific file system structures, all configurations expect the datasets to be located in a folder `/ssd`.
Therefore, you have to either mount the directory containing the dataset for the model to `/ssd` or create a symlink via:

```bash
ln -s asme/dataset/ /ssd
```

Also all output (checkpoints, logs, ...) will be written to `/output/...` bei default.

### Starting the training
You can find the training configs used to train our models inside the additional material archive.

Training of a model is started by executing:
```bash
asme train configs/training/<DATASET>/<MODEL>/config.jsonnet
```
By default, the training will be run on a single GPU. 
If you do not have a CUDA enabled GPU available, open `config.jsonnet` and change the `gpus` field to `0`. <br/>
You will have to adjust the `dirpath` field of the `checkpoint` object to a path where checkpoints can be saved.
Additionally, the `save_dir` field of the `tensorboard` object has to be changed to a valid path if you want to receive
logs during training. 
Alternatively, you can also symlink/mount the default paths to different directories.

## Evaluation
Again, this section assumes that you are located in the `asme` folder have already activated the virtual environment
via `poetry shell` (either locally or inside of the container)

### Evaluating a model
In order to evaluate the performance of a model, you need a model checkpoint generated during the training step. 
The checkpoints are saved in the directory specified in the config file used during training. <br>

You can only run an evaluation for a single target set size at a time so you need to adjust the original training config
accordingly by changing the property: `module.metrics.sampled.num_negative_samples` and `module.metrics.random_negative_samples.num_negative_samples`.

You can then evaluate the model checkpoint via
```bash
asme evaluate \
--config-file configs/training/<DATASET>/<MODEL>/config.jsonmet \
--checkpoint-file checkpoints/<MODEL_CHECKPOINT>.ckpt \
--output-file results/<DATASET>/<MODEL>/new-evaluate-<SEED>/results-<SAMPLE_SIZE>.json
--seed <SEED>
```
Note that the evaluation directory for each seed has to be named `new-evaluate-<SEED>` for technical reasons.

### Build metric results table 
After evaluating the models on different seeds, you can generate a table containing all results via
```bash
cd paper
python -m num_samples_experiment.build_results_table results results/results.csv
```
Where the first parameter points to the parent path where all results have been saved, .e.g the Path mounted to `/output` 
if a container is used.

### Build tables
```bash
cd paper
python -m true_rank_table.create_true_rank_table results/results.csv results/rank-table_recall-at-10_negative-samples --metric recall --k 10 --negative-samples 100
python -m true_rank_table.create_true_rank_table results/results.csv results/rank-table_ndcg-at-10_negative-samples --metric NDCG --k 10 --negative-samples 100
```

### Build target set size / rank plots
```bash
cd paper
METRIC="recall"
K=10
STRATEGY="pop_sampled"
#STRATEGY="uni_sampled"
DATASETS=( ml-1m ml-20m beauty games steam )
for DATASET in "${DATASETS[@]}" 
do
  python -m true_rank_table.create_rank_by_size_plot results/results.csv "results/rank_by_size/${DATASET}_${METRIC}-at-${K}_${STRATEGY}.pdf" $DATASET --metric $METRIC --k $K --sampling-strategy $STRATEGY
done
```

### Build strategy / rank plots
```bash
cd paper
METRIC="recall"
K=10
TGT_SIZE=100
DATASETS=( ml-1m ml-20m beauty games steam )
for DATASET in "${DATASETS[@]}" 
do
  python -m true_rank_table.create_dataset_rank_plot results/results.csv "results/rank_by_strategy/${DATASET}_${METRIC}-at-${K}_${TGT_SIZE}.pdf" --dataset $DATASET --metric $METRIC --k $K --negative-samples $TGT_SIZE
done
```
