import json
from pathlib import Path
from typing import Optional, TextIO

import re

import typer
import pandas as pd

NUM_SAMPLES_REGEX = re.compile(r".*?sampled\(([0-9]*?)\)")
AT_K_REGEX = re.compile(r"(.*?)@([0-9]+)")

POP_SAMPLED_METRIC_MODE=re.compile(r"(.*?)@[0-9]+?_sampled\([0-9]+\)")
UNIFORM_SAMPLED_METRIC_MODE=re.compile(r"(.*?)@[0-9]+?_random_negative_sampled\([0-9]+\)")


def determine_num_samples(metrics_name: str) -> Optional[int]:
    match = NUM_SAMPLES_REGEX.match(metrics_name)
    if match:
        return int(match.group(1))
    else:
        return 0


def determine_at_k_value(metric_name: str) -> int:
    match = AT_K_REGEX.match(metric_name)
    if match:
        return int(match.group(2))
    else:
        return 0


def get_metric_acronym(metric_name: str) -> str:
    if "@" in metric_name:
        return metric_name.split("@")[0]
    else:
        return metric_name


def get_sampling_mode(metric_name: str) -> str:
    pop_mode_match = POP_SAMPLED_METRIC_MODE.match(metric_name)
    if pop_mode_match:
        return "pop_sampled"

    uni_mode_match = UNIFORM_SAMPLED_METRIC_MODE.match(metric_name)
    if uni_mode_match:
        return "uni_sampled"

    return "exact"


def is_sampled_metric(metric_name: str) -> bool:
    return "sampled" in metric_name


def determine_seed(directory_name: str, evaluation_directory_prefix: str) -> Optional[int]:
    """
    Extracts the seed from the evaluation directory name. The directory name is expected to have a prefix
    `evaluation_directory_prefix` followed by a dash and the seed the seed, e.g. `evaluation-9383993`.

    :param directory_name: the name of the directory
    :return: a seed or 0 if no seed is encoded in the directory name.
    """
    if not directory_name.startswith(evaluation_directory_prefix):
        return None
    elif directory_name == evaluation_directory_prefix:
        return 0
    else:
        return int(directory_name.split("-")[2])


def sanitize_model_name(model_name: str) -> str:
    if "-" in model_name:
        return model_name.split("-")[0]

    if "_" in model_name:
        return model_name.split("_")[0]

    return model_name


def load_as_dataframe(file: TextIO) -> pd.DataFrame:
    results = json.load(file)
    metrics = results["metrics"]

    data = []
    for name, value in metrics.items():
        num_samples = determine_num_samples(name)
        metric_acronym = get_metric_acronym(name)
        sampling_mode = get_sampling_mode(name)
        k = determine_at_k_value(name)

        # we don't need validation losses in our table
        if "loss" in metric_acronym:
            continue

        data.append({
            "metric": metric_acronym,
            "k": k,
            "negative_samples": num_samples,
            "sampling_mode": sampling_mode,
            "value": float(value)
        })

    df = pd.DataFrame(data=data)
    df = df.astype({"k": "int32", "negative_samples": "int32"})

    return df


def collect_evaluation_results(eval_path: Path, source: str, model: str, dataset: str, evaluation_directory_prefix: str) -> pd.DataFrame:
    """
    Loads results for a single evaluation directory with multiple `results-*.json` files and aggregates everything into
    a DataFrame.
    :param eval_path: path to an evaluation directory, e.g. `<results>/<dataset_name>/<model_name>/evaluation`
    :param source: the source from which the results were collected, e.g. "asme" for results produced by asme.
    :param model: the name of the model that produced the results.
    :param dataset: the name of the dataset that was evaluated.

    :return: a DataFrame that aggregates over all results (seeds, metric sampling types).
    """
    df: pd.DataFrame = None
    for results_file in eval_path.glob("results-*.json"):
        with results_file.open("r") as file:
            if df is not None:
                df = pd.concat([df, load_as_dataframe(file)], ignore_index=True)
            else:
                df = load_as_dataframe(file)

    # every results file contains the exact results, so we're loading them repeatedly and need to drop the duplicate
    # rows.
    df = df.drop_duplicates(ignore_index=True)

    df["model"] = model
    df["dataset"] = dataset
    df["source"] = source
    df["seed"] = determine_seed(eval_path.name, evaluation_directory_prefix)

    return df


def get_dataset_paths(results_path: Path, evaluation_directory_prefix: str):
    """
    Collects all paths that contain results on the dataset level.

    :param results_path: a path that contains results for experiments on datasets.
    :param evaluation_directory_prefix: a prefix that indentifies directories that contain results.

    :return: a set of paths that point to results for datasets, e.g. `<result_path>/ml-1m`
    """
    model_evaluation_paths = results_path.glob(f"**/{evaluation_directory_prefix}*")
    dataset_paths = map(lambda p: p.parent.parent, model_evaluation_paths)
    return set(dataset_paths)


def main(results_path: Path,
         output_file: Path,
         delimiter: Optional[str] = "\t",
         evaluation_directory_prefix: Optional[str] = "new-evaluation"):

    dataset_paths = get_dataset_paths(results_path, evaluation_directory_prefix)

    df: pd.DataFrame = None

    for dataset_path in dataset_paths:

        if dataset_path.is_file():
            continue

        dataset_name = dataset_path.name
        evaluation_directories = list(dataset_path.glob(f"**/{evaluation_directory_prefix}*"))
        for evaluation_directory in evaluation_directories:
            if evaluation_directory.is_file():
                continue

            model_name = evaluation_directory.parent.name
            model_name = sanitize_model_name(model_name)

            collected_results = collect_evaluation_results(evaluation_directory, "asme", model_name, dataset_name, evaluation_directory_prefix)
            if df is not None:
                df = pd.concat([df, collected_results])
            else:
                df = collected_results

    # enforce some order to make the file more readable
    df.sort_values(by=["dataset", "model", "source", "metric", "sampling_mode", "k", "negative_samples"])
    rearranged_columns = ["source", "dataset", "model", "seed", "metric", "sampling_mode", "k", "negative_samples", "value"]
    df = df[rearranged_columns]
    df.reset_index(drop=True, inplace=True)

    df.to_csv(output_file, sep=delimiter)


if __name__ == "__main__":
    typer.run(main)
