import typer
import pandas as pd
from pathlib import Path

from plotnine import ggplot, aes, labs, geom_line, theme_bw, scale_x_log10, scale_y_log10, theme, element_blank, \
    element_text


app = typer.Typer()

@app.command(name="single")
def plot_single_config(input_file: Path, output_path: Path, dataset: str, metric: str, k: int, sampling_mode: str, seed: int, delimiter: str = "\t"):
    """
    Plots a metric value over the number of samples for a dataset and a fixed `k`.

    :param input_file: a csv file that contains all results.
    :param output_path: path where the output figure will be written.
    :param dataset: a dataset, can be one of: beauty, games, ml-1m, ml-20m, steam.
    :param metric: a metric, can be one of: mrr, recall, ndcg.
    :param k: the top-k considered items, can be one of: 1, 5, 10.
    :param sampling_mode: the sampling strategy used, can be one of: uni_sampled, pop_sampled,
    :param seed: a valid seed.
    :param delimiter: the delimiter used by the input file.

    """
    df = pd.read_csv(input_file, sep=delimiter)

    plot_data = df[(df.k == k) &
                   (df.metric == metric) &
                   (df.sampling_mode == sampling_mode) &
                   (df.dataset == dataset) &
                   (df.seed == seed)]

    if len(plot_data) == 0:
        print(f"The combination of parameters [k={k},metric={metric},sampling_mode={sampling_mode},dataset={dataset},seed={seed}] results in an empty table.")
        return

    fig = ggplot(plot_data, aes(x='negative_samples', y='value', color='model'))
    fig += labs(x=f"#samples", y=f"{metric}@{k}")
    fig += scale_x_log10()
    fig += geom_line(aes(linetype='model'))
    fig += theme_bw()
    fig += theme(legend_title=element_blank(), legend_position=(0.75, 0.7), legend_text=element_text(size=6))

    if not output_path.is_dir():
        print(f"{output_path} is not a directory.")
        return

    output_file_name = f"{dataset}-{metric}@{k}-{sampling_mode}-{seed}.pdf"

    fig.save(str(output_path / output_file_name), width=4, height=4)


@app.command("all")
def plot_all(input_file: Path, output_path: Path, delimiter: str = "\t"):
    """
    Takes a table and infers all possible combination between: seed, dataset, metric, k and sampling_mode, then plots it all.
    :param input_file: a csv file that contains all results.
    :param output_path: path where the output figures will be written.
    :param delimiter: the delimiter used by the input file.
    """
    if not output_path.is_dir():
        print(f"{output_path} is not a directory.")
        return

    df = pd.read_csv(input_file, sep=delimiter)

    datasets = df["dataset"].unique()
    metrics = df["metric"].unique()
    ks = df["k"].unique()
    sampling_modes = df["sampling_mode"].unique().tolist()
    # we need to exclude the true ranking since it does not change
    sampling_modes.remove("exact")

    seeds = df["seed"].unique()

    import itertools

    for dataset, metric, k, sampling_mode, seed in itertools.product(datasets, metrics, ks, sampling_modes, seeds):
        dataset_path = output_path / dataset
        metric_path = dataset_path / metric
        metric_path.mkdir(parents=True, exist_ok=True)

        try:
            plot_single_config(input_file, metric_path, dataset, metric, k, sampling_mode, seed, delimiter)
        except ValueError as error:
            print(f"Could not plot combination of parameters [k={k},metric={metric},sampling_mode={sampling_mode},dataset={dataset},seed={seed} due to an error: {error}")


if __name__ == "__main__":
    app()
