import itertools
import os
from pathlib import Path

import yaml
from zoller_python_utils.kubernetes import build_resource_requirements, build_env_def, build_secret_key_ref, \
    build_job_config, EmptyDirMemoryMount, escape_name, PrioClass

from paper.common import DATASETS, MODELS, ROOT_PATH, Dataset, Model
from paper.k8s.k8s_zoller_constants import SCRATCH, USER_HOME

ASME_IMAGE_URL = 'gitlab2.informatik.uni-wuerzburg.de:4567/dmir/dallmann/recommender/asme-dev:latest'

PAPER_CODE_PATH = '/home/ls6/zoller/research/sampling-paper'

CACHE = EmptyDirMemoryMount('cache', '/ssd')

PRIORITY = PrioClass.LOW


def _build_exact_rank_config(dataset: Dataset,
                             model: Model):
    dataset_sub_path = dataset.sub_path
    dataset_path = ROOT_PATH / dataset_sub_path

    envs = [
        build_env_def('PYTHONPATH', python_path),
        build_env_def('PROJECT_DIR', str(path_to_project)),
        build_env_def('REPO_USER', 'daz65be'),
        build_env_def('REPO_BRANCH', 'master'),
        build_env_def('GIT_TOKEN', value_from=secret_ref),
        build_env_def('PREPARE_SCRIPT',
                      f'/home/ls6/zoller/research/sampling-paper/k8s/training/{dataset_sub_path}/copy_to_cache.sh')
    ]

    model_name = model.name.lower()

    output_folder = Path(f'../../../k8s/rank-experiment/{dataset_sub_path}/{model_name}')
    os.makedirs(output_folder, exist_ok=True)

    model_path = dataset_path / model.sub_path
    checkpoint_file = model_path / 'checkpoints' / 'best.ckpt'
    config_file = model_path / 'config_rank_experiment.json'
    args = ['predict',
            f'{model_path / "best_predictions.json"}',
            f'--checkpoint-file={checkpoint_file}',
            f'--config-file={config_file}',
            f'--gpu=1'
            ]

    job_config = build_job_config(escape_name(f'{dataset.normed_name()}-{model_name}-predict'),
                                  'worker', ASME_IMAGE_URL,
                                  [USER_HOME, SCRATCH, CACHE],
                                  args=args,
                                  requests=resources,
                                  envs=envs)
    with open(output_folder / 'exact.yml', 'w') as outfile:
        yaml.dump(job_config, outfile, default_flow_style=False)


if __name__ == '__main__':
    num_gpus = 1
    resources = build_resource_requirements(cpu_count=12, memory=24, gpu_count=num_gpus)
    path_to_project = Path('/home/ls6/zoller/seq_recommendation/project_master')

    secret_ref = build_secret_key_ref('token', 'gitlab-token')
    python_path = str(path_to_project / 'recommender/src')

    for dataset, model in itertools.product(DATASETS, MODELS):
        _build_exact_rank_config(dataset, model)
        for num_examples in dataset.num_examples:
            dataset_name = dataset.name
            dataset_sub_path = dataset.sub_path
            dataset_path = ROOT_PATH / dataset_sub_path

            envs = [
                build_env_def('PYTHONPATH', PAPER_CODE_PATH),
                build_env_def('PROJECT_DIR', str(path_to_project)),
                build_env_def('REPO_USER', 'daz65be'),
                build_env_def('REPO_BRANCH', 'master'),
                build_env_def('GIT_TOKEN', value_from=secret_ref),
                build_env_def('PREPARE_SCRIPT',
                              f'{PAPER_CODE_PATH}/k8s/training/{dataset_sub_path}/copy_to_cache.sh')
            ]

            model_name = model.name.lower()

            output_folder = Path(f'../../../k8s/rank-experiment/{dataset_sub_path}/{model_name}')
            os.makedirs(output_folder, exist_ok=True)

            model_path = dataset_path / model.sub_path
            checkpoint_file = model_path / 'checkpoints' / 'best.ckpt'
            config_file = model_path / 'config_rank_experiment.json'
            command = ['python']
            args = [f'{PAPER_CODE_PATH}/paper/rank_experiment/k8s/run_ranking_sampled_job.py',
                    f'{dataset.name}',
                    f'{model.name}',
                    f'{num_examples}'
                    ]

            job_config = build_job_config(escape_name(f'{dataset.normed_name()}-{model_name}-{num_examples}-predict'),
                                          'worker', ASME_IMAGE_URL,
                                          [USER_HOME, SCRATCH, CACHE],
                                          command=command,
                                          args=args,
                                          requests=resources,
                                          envs=envs,
                                          prio_class=PRIORITY)
            with open(output_folder / f'{num_examples}.yml', 'w') as outfile:
                yaml.dump(job_config, outfile, default_flow_style=False)