import json
import subprocess
from io import StringIO
from pathlib import Path

import typer

from paper.common import get_dataset_by_name, SEEDS, get_model_by_name, ROOT_PATH


def _update_config(old_config_path: Path,
                   new_config_path: Path):
    with open(old_config_path) as old_config_file, open(new_config_path, 'w') as new_config_file:
        config = json.load(old_config_file)
        config['data_sources']['test']['loader']['num_workers'] = 12
        config['module']['metrics']['full']['metrics']['rank'] = []
        config['module']['metrics']['random_negative_sampled']['metrics']['rank'] = []
        config['module']['metrics']['sampled']['metrics']['rank'] = []
        trainer_config = config['trainer']
        log_path = str(new_config_path.parent / 'predict_logs')
        trainer_config['checkpoint']['dirpath'] = log_path
        trainer_config['loggers']['tensorboard']['save_dir'] = log_path

        json.dump(config, new_config_file)


def main(dataset_name: str,
         model_name: str,
         num_examples: int):
    # first update config to include the rank metric
    print(f'running {model_name} for {dataset_name} and {num_examples}')

    dataset = get_dataset_by_name(dataset_name)
    model = get_model_by_name(model_name)

    model_output = ROOT_PATH / dataset.sub_path / model.sub_path

    for seed in SEEDS:
        print(f'\trunning seed {seed}')
        seed_result_path = model_output / f'new-evaluation-{seed}'
        config_path = seed_result_path / f'config-{num_examples}_rank_experiment.jsonnet'

        print('\t\tupdating run config')
        old_config = seed_result_path / f'config-{num_examples}.jsonnet'
        _update_config(old_config, config_path)

        print('\t\trunning predict')
        program = ["/entrypoint.sh", 'predict', f'{seed_result_path}/best_predictions_{num_examples}.json',
                f'--checkpoint-file={model_output}/checkpoints/best.ckpt',
                f'--config-file={config_path}', f'--seed={seed}', f'--gpu=1', '--overwrite']
        print(program)
        result = subprocess.Popen(program)
        output, error = result.communicate()
        print(StringIO(output))
        print(StringIO(error))
        return_code = result.returncode

        if return_code != 0:
            raise ValueError('run not successful')


if __name__ == "__main__":
    typer.run(main)
