import argparse
import json
import os
from pathlib import Path
from typing import Dict

import pandas as pd
from plotnine import ggplot, aes, labs, geom_line, theme_bw, scale_x_log10, scale_y_log10, theme, element_blank, \
    element_text

from paper.common import get_dataset_by_name, MODELS, Model, Dataset, ROOT_PATH


def _load_rank_distribution(ranks_file: Path
                            ) -> Dict[int, int]:
    with open(ranks_file) as ranks:
        return {
            int(line.split(' ')[0]): int(line.split(' ')[1]) for line in ranks.readlines()
        }


def _write_rank_distribution(ranks_file_path: Path,
                             rank_distribution: Dict[int, int]
                             ) -> None:
    with open(ranks_file_path, 'w') as rank_file:
        for rank, count in rank_distribution.items():
            rank_file.write(f'{rank} {count}\n')


def _get_rank_distribution_of_model(dataset: Dataset,
                                    model: Model
                                    ) -> Dict[int, int]:
    model_path = ROOT_PATH / dataset.sub_path / model.sub_path
    ranks_file_path = model_path / 'ranks.txt'
    if os.path.exists(ranks_file_path):
        return _load_rank_distribution(ranks_file_path)

    predictions_file_path = model_path / 'best_predictions.json'

    with open(predictions_file_path) as predictions_file:
        ranks = {}

        print('counting ranks')
        for line in predictions_file.readlines():
            results = json.loads(line)
            rank = results['metrics']['rank']
            current_count = ranks.get(rank, 0)
            current_count += 1
            ranks[rank] = current_count

        _write_rank_distribution(ranks_file_path, ranks)
        return ranks


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='most popular for location')
    parser.add_argument('dataset', type=str, help='dataset name')

    args = parser.parse_args()

    dataset = get_dataset_by_name(args.dataset)

    values = []

    for model in MODELS:
        distribution = _get_rank_distribution_of_model(dataset, model)
        for rank, count in distribution.items():
            values.append((model.name, rank, count))

    data = pd.DataFrame(values, columns=['model', 'rank', 'count'])

    fig = ggplot(data, aes(x='rank', y='count', color='model'))
    fig += labs(x="Rank", y="Frequency")
    fig += scale_x_log10()
    fig += scale_y_log10()
    fig += geom_line(aes(linetype='model'))
    fig += theme_bw()
    fig += theme(legend_title=element_blank(), legend_position=(0.75, 0.7), legend_text=element_text(size=6))

    fig.save(f'{dataset.normed_name()}.pdf', width=4, height=4)
