from typing import List, Optional

import pandas as pd

RANDOM_SAMPLING_MODES = ["pop_sampled", "uni_sampled"]

EXACT_MODE = "exact"

SAMPLING_MODES = RANDOM_SAMPLING_MODES + [EXACT_MODE]

DATASETS = ["beauty", "games", "ml-1m", "ml-20m", "steam"]


def filter_data(df: pd.DataFrame,
                dataset: str,
                metric: str,
                k: int,
                negative_samples: Optional[int] = None,
                sampling_mode: Optional[str] = None,
                include_exact_strategy: Optional[bool] = True,
                min_number_negative_samples: Optional[int] = None
                ) -> pd.DataFrame:
    """
    Selects only rows for the given dataset, metric, k and negative_samples.

    :param df: a DataFrame.
    :param dataset: a dataset.
    :param metric: a metric.
    :param k: the cut-off parameter for the metric.
    :param negative_samples: the number of negative samples drawn during evaluation.
    :return:
    """
    filtered = df[(df.dataset == dataset) &
                  (df.k == k) &
                  (df.metric == metric)]

    if sampling_mode is not None:
        filtered = filtered[(filtered.sampling_mode == sampling_mode)]

    if negative_samples is not None:
        if include_exact_strategy:
            filtered = filtered[(filtered.negative_samples == negative_samples) | (filtered.negative_samples == 0)]
        else:
            filtered = filtered[(filtered.negative_samples == negative_samples)]

    if min_number_negative_samples is not None:
        filtered = filtered[(filtered.negative_samples >= min_number_negative_samples)]

    return filtered.sort_values(by=["model", "value"])


def calc_mean_value(df: pd.DataFrame) -> pd.DataFrame:
    """
    For every sampling mode, calculates the mean and std deviation of the metric value across all seeds and adds them as
     a column to the DataFrame.

    :param df: a DataFrame
    :return: a DataFrame that includes mean and standard deviation for the value computed across all seeds.
    """
    df = df[["model", "value", "sampling_mode", "seed"]]
    gb = df.groupby(["model", "sampling_mode"])
    df = gb.size().to_frame(name="test")
    df = df.join(gb.agg({"value": "mean"}).rename(columns={"value": "mean"})) \
        .join(gb.agg({"value": "std"}).rename(columns={"value": "std"})).reset_index()
    df = df[["model", "sampling_mode", "mean", "std"]] #remove test column
    return df


def calc_mean_value_by_negative_samples(df: pd.DataFrame) -> pd.DataFrame:
    df = df[["model", "value", "negative_samples", "sampling_mode", "seed"]]
    gb = df.groupby(["model", "sampling_mode", "negative_samples"])
    df = gb.size().to_frame(name="test")
    df = df.join(gb.agg({"value": "mean"}).rename(columns={"value": "mean"})) \
        .join(gb.agg({"value": "std"}).rename(columns={"value": "std"})).reset_index()
    df = df[["model", "sampling_mode", "negative_samples", "mean", "std"]] #remove test column

    return df


def add_ranks(df: pd.DataFrame, sampling_strategies: List[str]) -> pd.DataFrame:
    """
    For every sampling strategy, ranks the models according to the metric value.
    :param df: a DataFrame
    :param sampling_strategies: a list of sampling strategies to consider.

    :return: a DataFrame that includes a columnt with ranks for every model.
    """
    ranked_table = None
    for sampling_strategy in sampling_strategies:  #FIXME: determine sampling strategies from table
        result_table = df[df.sampling_mode == sampling_strategy].sort_values(by="mean", ascending=False)
        result_table["rank"] = [1, 2, 3, 4]  #FIXME: determine number of needed ranks from models included in table

        if ranked_table is None:
            ranked_table = result_table
        else:
            ranked_table = ranked_table.append(result_table)

    return ranked_table.sort_values(by=["sampling_mode", "model"])


def add_ranks_by_negative_samples(df: pd.DataFrame, sampling_strategy: str) -> pd.DataFrame:
    # determine negative_sample values
    negative_samples_values = df["negative_samples"].unique().tolist()

    ranked_table = None
    for negative_samples in negative_samples_values:
        result_table = df[(df.sampling_mode == sampling_strategy) & (df.negative_samples == negative_samples)].sort_values(by="mean", ascending=False)
        result_table["rank"] = [1, 2, 3, 4]  #FIXME: determine number of needed ranks from models included in table

        if ranked_table is None:
            ranked_table = result_table
        else:
            ranked_table = ranked_table.append(result_table)

    ranked_table = ranked_table.sort_values(by=["negative_samples", "model", "rank"])
    return ranked_table


def preprocess_table(df: pd.DataFrame,
                     dataset: str,
                     metric: str,
                     k: int,
                     negative_samples: int,
                     sampling_strategies: List[str] = ["exact", "pop_sampled", "uni_sampled"]) -> pd.DataFrame:
    result = filter_data(df, dataset, metric, k, negative_samples)
    result = calc_mean_value(result)
    return add_ranks(result, sampling_strategies)


def preprocess_table_by_negative_samples(df: pd.DataFrame, dataset: str,
                                         metric: str,
                                         k: int,
                                         sampling_strategy: str) -> pd.DataFrame:
    result = filter_data(df, dataset, metric, k, sampling_mode=sampling_strategy, include_exact_strategy=False, min_number_negative_samples=100)
    result = calc_mean_value_by_negative_samples(result)
    return add_ranks_by_negative_samples(result, sampling_strategy)