import os
from pathlib import Path
from typing import Optional

import pandas as pd
import typer

from true_rank_table.common import preprocess_table, SAMPLING_MODES, DATASETS
from typing import List, Dict

from plotnine import *


SAMPLING_MODE_TO_NUMERIC = {
    "exact": 2,
    "pop_sampled": 3,
    "uni_sampled": 1
}


SAMPLE_XLABELS = {
    2: "Full",
    3: "Popular",
    1: "Uniform"
}


def _get_rank_labels(df: pd.DataFrame) -> List[str]:
    return list(sorted(df["rank"].unique().tolist()))


def _get_x_labels(mapping: Dict[str, int]) -> List[str]:
    return list(map(lambda x: SAMPLE_XLABELS[x], sorted(list(mapping.values()))))


def plot(input_file: Path,
         output_path: Path,
         dataset: str = None,
         metric: str = "recall",
         k: int = 10,
         negative_samples: int = 100,
         delimiter: Optional[str] = "\t"):

    df = pd.read_csv(input_file, sep=delimiter, index_col=0)

    exp_ids = [dataset] if dataset is not None else DATASETS

    if not output_path.exists():
        os.makedirs(output_path)

    for dataset_experiment_id in exp_ids:
        dataset_df = preprocess_table(df, dataset_experiment_id, metric, k, negative_samples, SAMPLING_MODES)
        for sampling_mode, id in SAMPLING_MODE_TO_NUMERIC.items():
            dataset_df.loc[dataset_df["sampling_mode"] == sampling_mode, "sampling_mode"] = id
        dataset_df["sampling_mode"] = pd.to_numeric(dataset_df["sampling_mode"])

        theme_set(theme_void())

        print(dataset_df)

        rank_aes = aes(x="sampling_mode", y="rank", color="model", linetype='model')

        fig = ggplot(dataset_df, rank_aes)

        # add some lines for each rank
        rank_labels = _get_rank_labels(dataset_df)
        for rank_label in rank_labels:
            fig += geom_hline(yintercept=rank_label, linetype="dotted", color="lightgrey", size=1.0)

        fig += geom_point(rank_aes, size=7.5)
        fig += geom_line(rank_aes, size=2.5)
        fig += labs(y="Rank", x="")

        x_labels = _get_x_labels(SAMPLING_MODE_TO_NUMERIC)
        fig += scale_y_discrete(limits=rank_labels)
        fig += scale_y_reverse()
        fig += scale_x_continuous(labels=x_labels, breaks=[1, 2, 3])

        fig += theme(plot_title=element_text(dataset),
                     legend_title=element_blank(),
                     legend_position='none',
                     legend_text=element_text(size=20),
                     axis_title=element_text(size=28),
                     legend_background=element_rect(colour='lightgrey', size=0.5),
                     text=element_text(color='#222222'),
                     rect=element_rect(fill='#FFFFFF'),
                     axis_text=element_text(size=28),
                     axis_title_y=element_text(size=28, rotation=90, margin={"r": 15.0}),
                     axis_text_x=element_text(margin={'t': 15.0}),
                     aspect_ratio=1/2)
        fig.save(output_path / f'{dataset_experiment_id}_{metric}-at-{k}_samples-{negative_samples}.pdf')

    # this generates the legend "plot"
    # this is a complete hack because there is no option in nineplot to only generate the legend
    legend_figure = build_legend_plot(dataset_df, ['Models', 'sampling_mode', 'mean', 'std', 'rank'])
    legend_figure.save(output_path / 'legends.pdf', width=1, height=1)


def build_legend_plot(dataset_df, column_names):
    rank_aes = aes(x="sampling_mode", y="rank", color="Models", linetype='Models')
    print(dataset_df.columns)
    dataset_df.columns = column_names
    dataset_df['Models'] = dataset_df['Models'].str.replace('bert4rec', 'BERT4Rec')
    dataset_df['Models'] = dataset_df['Models'].str.replace('gru', 'GRU')
    dataset_df['Models'] = dataset_df['Models'].str.replace('narm', 'NARM')
    dataset_df['Models'] = dataset_df['Models'].str.replace('sasrec', 'SASRec')
    leg_fig = ggplot(dataset_df, rank_aes)
    leg_fig += theme_void()
    leg_fig += geom_line(size=2.5)
    leg_fig += geom_hline(yintercept=2, color="white", size=50.0)
    leg_fig += theme(legend_title=element_text(size=20),
                     legend_text=element_text(size=20),
                     legend_position='left',
                     legend_spacing=100,
                     legend_key_width=50)
    return leg_fig


if __name__ == "__main__":
    typer.run(plot)
