import itertools
import os
from pathlib import Path
from typing import Optional, List

import pandas as pd
import typer

from true_rank_table.common import preprocess_table_by_negative_samples, DATASETS, RANDOM_SAMPLING_MODES, \
    preprocess_table, EXACT_MODE

from plotnine import *

from true_rank_table.create_dataset_rank_plot import build_legend_plot

RANDOM_SAMPLING_MODES_NAME = {
    'pop_sampled': 'Randomly Sampled by Popularity',
    'uni_sampled': 'Uniform Random Sampled'
}


class SampleSizeMapping:
    def __init__(self, df: pd.DataFrame):
        self.sample_sizes = sorted(list(df["negative_samples"].unique().tolist()))

    def add_equidistant_numeric_mapping(self, df: pd.DataFrame):
        for id, negative_sample_size in enumerate(self.sample_sizes, start=1):
            df.loc[df["negative_samples"] == negative_sample_size, "negative_samples_equi"] = id
        df["negative_samples_equi"] = pd.to_numeric(df["negative_samples_equi"])
        return df


def _get_rank_labels(df: pd.DataFrame) -> List[str]:
    return list(sorted(df["rank"].unique().tolist()))


def plot(input_file: Path,
         output_path: Path,
         dataset: str = None,
         metric: str = "recall",
         k: int = 10,
         sampling_strategy: str = None,
         delimiter: Optional[str] = "\t"):

    datasets = [dataset] if dataset is not None else DATASETS
    sampling_strategies = [sampling_strategy] if sampling_strategy is not None else RANDOM_SAMPLING_MODES

    if not output_path.exists():
        os.makedirs(output_path)

    df = pd.read_csv(input_file, sep=delimiter, index_col=0)

    for dataset, sampling_strategy in itertools.product(datasets, sampling_strategies):
        preprocessed_df = preprocess_table_by_negative_samples(df, dataset, metric, k, sampling_strategy)
        sample_size_mapping = SampleSizeMapping(preprocessed_df)
        preprocessed_df = sample_size_mapping.add_equidistant_numeric_mapping(preprocessed_df)
        #print(preprocessed_df)

        exact_ranking_df = preprocess_table(df, dataset, metric, k, 0, [EXACT_MODE])
        full_id = max(preprocessed_df['negative_samples_equi'].unique().tolist()) + 1
        exact_ranking_df['negative_samples_equi'] = full_id
        exact_ranking_df['negative_samples'] = 'full'

        preprocessed_df = preprocessed_df.append(exact_ranking_df)

        print(preprocessed_df)

        theme_set(theme_void())
        rank_aes = aes(x="negative_samples_equi", y="rank", color="model", linetype="model")
        fig = ggplot(preprocessed_df, rank_aes)

        rank_labels = _get_rank_labels(preprocessed_df)
        for rank_label in rank_labels:
            fig += geom_hline(yintercept=rank_label, linetype="dotted", color="lightgrey")

        negative_sample_ids = preprocessed_df["negative_samples"].unique().tolist()

        x_labels = list(map(lambda x: str(x), negative_sample_ids))
        fig += scale_x_discrete(limits=x_labels, labels=x_labels)
        fig += geom_point(rank_aes, size=3.0)
        fig += geom_line(size=1.5)
        fig += labs(x=f"Number of Negative Samples", y=f"Rank")

        fig += scale_y_discrete(limits=rank_labels)
        fig += scale_y_reverse()

        fig += ggtitle(RANDOM_SAMPLING_MODES_NAME[sampling_strategy])
        fig += theme(legend_position='none',
                     axis_title=element_text(size=12),
                     text=element_text(color='#222222'),
                     rect=element_rect(fill='#FFFFFF'),
                     axis_text=element_text(size=10),
                     axis_title_y=element_text(rotation=90, margin={"r": 10.0}),
                     axis_text_x=element_text(margin={'t': 15.0}),
                     axis_title_x=element_text(margin={'t': 15.0}),
                     aspect_ratio=1/4.5)
        fig.save(output_path / f'{dataset}_{metric}-at-{k}_{sampling_strategy}.pdf')

    legend_plot = build_legend_plot(preprocessed_df, ['Models', 'sampling_mode', 'negative_samples', 'mean', 'std', 'rank',
                                                      'negative_samples_equi'])
    legend_plot += theme(legend_position='top')
    legend_plot.save(output_path / 'legends.pdf', width=1, height=1)


if __name__ == "__main__":
    typer.run(plot)
