from pathlib import Path
from typing import Optional, List, Any

from true_rank_table.common import SAMPLING_MODES, DATASETS, preprocess_table

import pandas as pd
import typer

MODELS = ["gru", "narm", "sasrec", "bert4rec"]

DATASET_TO_COMMAND = {
    "beauty": "\\ambeauty",
    "games": "\\amgame",
    "steam": "\\steam",
    "ml-1m": "\\movielens",
    "ml-20m": "\\movielenslarge"

}

MODEL_TO_COMMAND = {
    "gru": "\\gru",
    "narm": "\\narm",
    "sasrec": "\\sasrec",
    "bert4rec": "\\bertforrec"
}



def get_header(metric: str, k: int) -> str:
    if metric == "recall":
        metric_command = f"\\hr{{{k}}}"
    if metric == "NDCG":
        metric_command = f"\\ndcg{{{k}}}"

    return f"""
    \\begin{{tabular}}{{llrlrlcrlc}}
    \\toprule
 {{}} & {{}}      & \multicolumn{{2}}{{c}}{{true}} & \multicolumn{{3}}{{c}}{{by popularity}} & \multicolumn{{3}}{{c}}{{uniform random}} \\\\
 Dataset   & Model   & {metric_command}  & rank           & {metric}@{k}  & rank  & $\\tau$        & {metric_command}  & rank & $\\tau$ \\\\
    \\midrule
"""


def get_cell_value(df: pd.DataFrame, model: str, sampling_mode: str, column_name: str) -> Any:
    return df[(df.model == model) & (df.sampling_mode == sampling_mode)][column_name].item()


def decorate_value_with_rank(value_markup: str, rank: int) -> str:
    decorated_markup = value_markup
    if rank == 1:
        decorated_markup = f"\\textbf{{{value_markup}}}"
    if rank == 2:
        decorated_markup = f"\\underline{{{value_markup}}}"

    return decorated_markup + " "


def get_body(df: pd.DataFrame, dataset: str, pop_tau: float, uni_tau: float, models: List[str], sampling_modes: List[str]) -> str:
    rows = []
    tau_added = False

    for model in models:
        row = []
        row.append(f"    {MODEL_TO_COMMAND[model]} ")

        for sampling_mode in sampling_modes:
            rank = get_cell_value(df, model, sampling_mode, 'rank')

            value_markup = f"\\num[round-mode=places,round-precision=3]{{{get_cell_value(df, model, sampling_mode, 'mean')}}}"
            row.append(decorate_value_with_rank(value_markup, rank))

            single_square = "\\blacksquare"
            squares = f"${rank * single_square}$"
            row.append(f"{squares} ")

            if sampling_mode == "pop_sampled" or sampling_mode == "uni_sampled":
                if sampling_mode == "pop_sampled":
                    tau = pop_tau
                if sampling_mode == "uni_sampled":
                    tau = uni_tau

                if not tau_added:
                    row.append(f"\multirow{{4}}{{*}}{{\\num[round-mode=places,round-precision=2]{{{tau}}}}} ")

                    if sampling_mode == "uni_sampled":
                        tau_added = True
                else:
                    row.append(" ")

        row_as_str = " & " + " & ".join(row)
        rows.append(row_as_str)
    return f"\multirow{{4}}{{*}}{{{DATASET_TO_COMMAND[dataset]}}} " + "\\\\\n".join(rows) + "\\\\"


def get_footer() -> str:
    return """\\bottomrule
    \\end{tabular}
    """


def create_table(input_file: Path, output_file: Path, metric: str = "recall", k: int = 10, negative_samples: int = 100, delimiter: Optional[str] = "\t"):
    df = pd.read_csv(input_file, sep=delimiter, index_col=0)

    output = get_header(metric, k)

    for idx, dataset in enumerate(DATASETS):
        dataset_df = preprocess_table(df, dataset, metric, k, negative_samples)
        exact_ranks = dataset_df[(dataset_df["sampling_mode"] == "exact")]["rank"]
        pop_ranks = dataset_df[(dataset_df["sampling_mode"] == "pop_sampled")]["rank"]
        uni_ranks = dataset_df[(dataset_df["sampling_mode"] == "uni_sampled")]["rank"]
        from scipy.stats import kendalltau
        pop_exact_tau = kendalltau(x=exact_ranks, y=pop_ranks)
        uni_exact_tau = kendalltau(x=exact_ranks, y=uni_ranks)
        print(f"{dataset}  tau_pop = {pop_exact_tau}")
        print(f"{dataset}  tau_uni = {uni_exact_tau}")
        # get kendalls tau
        output += get_body(dataset_df, dataset, pop_exact_tau.correlation, uni_exact_tau.correlation, MODELS, SAMPLING_MODES)

        if idx < len(DATASETS) - 1:
            output += "\n\\midrule"
        output += "\n"
    output += get_footer()

    with output_file.open("w") as of:
        of.write(output)


if __name__ == "__main__":
    typer.run(create_table)
